import logging

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from formica.settings import app_config
from formica.web.api_routes.main import api_router
from formica.web.web_routes import web_router
from starlette.staticfiles import StaticFiles

app = FastAPI()
logger = logging.getLogger(__name__)

origins = [
    "*",
]

app.mount(
    "/assets", StaticFiles(directory="src/formica/web/dist/assets"), name="assets"
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(api_router, prefix="/api")
app.include_router(web_router, prefix="")


def run_webserver():
    import uvicorn

    uvicorn.run(
        "formica.web.main:app",
        host=app_config.get("webserver", "HOST"),
        port=app_config.getint("webserver", "PORT"),
        reload=False,
    )


if __name__ == "__main__":
    run_webserver()



import re
from typing import Optional

from pydantic import BaseModel


class Token(BaseModel):
    access_token: str
    token_type: str


class TokenData(BaseModel):
    username: Optional[str] = None


def camel_to_snake(name):
    # Converts camelCase or PascalCase to snake_case
    s1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def convert_keys_to_snake_case(obj):
    if isinstance(obj, dict):
        return {
            camel_to_snake(k): convert_keys_to_snake_case(v) for k, v in obj.items()
        }
    elif isinstance(obj, list):
        return [convert_keys_to_snake_case(i) for i in obj]
    else:
        return obj



import contextlib
from functools import wraps
from inspect import signature
from typing import Any
from typing import AsyncGenerator
from typing import Awaitable
from typing import Callable
from typing import cast
from typing import ParamSpec
from typing import TypeVar

from formica.settings import engine
from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.orm import sessionmaker
from sqlmodel.ext.asyncio.session import AsyncSession
from sqlmodel.ext.asyncio.session import AsyncSession as SASession

SessionLocal = async_sessionmaker(bind=engine, class_=SASession, expire_on_commit=False)


async def get_session() -> AsyncGenerator[Any, Any]:
    async_session = sessionmaker(engine, class_=AsyncSession, expire_on_commit=False)
    async with async_session() as session:
        yield session


@contextlib.asynccontextmanager
async def _create_session() -> AsyncGenerator[AsyncSession, None]:
    """Contextmanager that will create and teardown a session."""
    async with SessionLocal() as session:
        try:
            yield session
            await session.commit()
        except Exception:
            await session.rollback()
            raise


PS = ParamSpec("PS")
RT = TypeVar("RT")


def find_session_idx(func: Callable[PS, RT]) -> int:
    """Find the index of the 'session' parameter in the function signature."""
    func_params = signature(func).parameters
    try:
        return tuple(func_params).index("session")
    except ValueError:
        raise ValueError(
            f"Function {func.__qualname__} has no 'session' parameter"
        ) from None


def provide_session(func: Callable[PS, Awaitable[RT]]) -> Callable[PS, Awaitable[RT]]:
    """
    Async decorator to provide a session to a coroutine if not provided.
    """
    session_args_idx = find_session_idx(func)

    @wraps(func)
    async def wrapper(*args, **kwargs) -> RT:
        if "session" in kwargs or session_args_idx < len(args):
            return await func(*args, **kwargs)
        else:
            async with _create_session() as session:
                return await func(*args, session=session, **kwargs)

    return wrapper


# A fake session to use in functions decorated by provide_session. This allows
# the 'session' argument to be of type Session instead of Session | None,
# making it easier to type hint the function body without dealing with the None
# case that can never happen at runtime.
NEW_SESSION: SASession = cast(SASession, None)



import os

from fastapi import APIRouter
from starlette.responses import FileResponse

web_router = APIRouter()


# Serve the index.html for the root path and all unmatched routes
@web_router.get("/{full_path:path}")
async def serve_next(full_path: str):
    # If the path exists as a file, serve it directly
    file_path = os.path.join("src/formica/web/dist", full_path)
    if os.path.exists(file_path) and os.path.isfile(file_path):
        return FileResponse(file_path)

    # Otherwise return index.html to let React handle the routing
    return FileResponse("src/formica/web/dist/index.html")



from __future__ import annotations

import copy

from formica.models.models import FlowVersionModel
from formica.node.nodes import NodeFactory
from formica.utils.validate import convert_keys_to_snake_case


class Flow:
    def __init__(self, flow_version: FlowVersionModel, arguments=None):
        """
        Khởi tạo Flow từ cấu trúc Flow dạng JSON, khởi tạo các Node thêm vào node_dict
        Set các đối số được truyền vào từ `arguments`
        :param flow_id: flow id
        :param structure: cấu trúc đồ thị của Flow
        :param arguments: các đối số truyền vào

        :raises ValueError: Tên đối số không tồn tại trong danh sách tham số của Flow
        """
        if arguments is None:
            arguments = {}
        self.flow_id = flow_version.flow_id
        self.node_dict = {}
        self.params: list[str] = flow_version.parameters
        self.vars: dict[str, str] = {}
        structure_ = copy.deepcopy(flow_version.structure)

        self._deserialize(structure_)
        try:
            self._pass_arguments(arguments)
        except ValueError as e:
            raise e

    def _pass_arguments(self, arguments: dict[str, str]):
        """
        Gọi hàm này để truyền các đối số (arguments) vào flow
        Các giá trị này sau đó sẽ được set vào self.vars (global giữa các toán tử)
        :param arguments: các đối số
        :return:

        :raises ValueError: Tên đối số không tồn tại trong danh sách tham số của Flow
        """
        for arg_name in arguments:
            if arg_name not in self.params:
                raise ValueError(
                    f"Tên đối số không tồn tại trong danh sách tham số của Flow: {arg_name}"
                )

        # Tới đây được nghĩa là tất cả đối số đã hợp lệ
        self.vars.update(arguments)

    def _deserialize(self, structure: dict) -> None:
        """
        Parse dạng json của flow ra thành các toán tử liên kết với nhau trong node_dict

        Đồng thời cũng parse các tham số truyền vào và biến global
        :param structure: Dạng json của flow
        :return: None
        """
        flow_json = structure
        # version = flow_json["version"]
        # if version == "0.1":
        self._parse_flow_v0_1(flow_json)
        # else:
        #     raise ValueError(f"Unknown flow version {version}")

    def _parse_flow_v0_1(self, flow_json: dict):
        # Parse các tham số (biến global)
        # self.vars = flow_json["globals"]
        # Parse các Node và đẩy vào `node_dict`
        flow_json = self._preprocess_flow(flow_json)
        # print(flow_json)
        for node_json in flow_json["nodes"]:
            new_node = NodeFactory.create_node(node_json)
            new_node.flow = self
            self.node_dict[new_node.node_id] = new_node

        # Update the edges
        for edge_json in flow_json["edges"]:
            print(edge_json)
            print("+=============================================================")
            self.node_dict[edge_json["source"]].add_downstream(
                self.node_dict[edge_json["target"]]
            )

    def _preprocess_flow(self, flow_json: dict) -> dict:
        flow_json = convert_keys_to_snake_case(flow_json)
        for node in flow_json["nodes"]:
            if node["data"]["label"] == "Decision":
                for edge in flow_json["edges"]:
                    if (
                        edge["source"] == node["id"]
                        and edge["source_handle"] == "out-1"
                    ):
                        node["data"]["config"]["true_branch"] = edge["target"]
                    elif (
                        edge["source"] == node["id"]
                        and edge["source_handle"] == "out-2"
                    ):
                        node["data"]["config"]["false_branch"] = edge["target"]

        return flow_json

    def print_flow(self):
        """Hàm này để debug"""
        for node in self.node_dict.values():
            print(f"{node.node_id} => {node.downstream_node_ids}")



import logging
import traceback

from formica.execution.context import Context
from formica.models.constant import TaskState
from formica.models.constant import TaskTriggerRule
from formica.node.nodes import BaseNode

logger = logging.getLogger(__name__)


class Task:
    def __init__(self, node: BaseNode, trigger_rule: TaskTriggerRule):
        self.node: BaseNode = node
        self.state = TaskState.WAIT_FOR_EXECUTING
        self.trigger_rule = trigger_rule

    async def run(self, context: Context) -> None:
        self.state = TaskState.RUNNING
        try:
            await self.node.execute(context)
            self.state = TaskState.SUCCESS
        except Exception as e:
            logger.warning(
                f"Error while executing task {self.node.node_id} (type {self.node.__class__}): {e}"
            )
            traceback.print_exc()

            self.state = TaskState.FAILED



import asyncio
import logging
import time
from time import sleep
from typing import Optional

from formica.executor.executor import Executor
from formica.executor.executor import LocalExecutor
from formica.models.constant import DeviceRunState
from formica.models.constant import FlowRunState
from formica.models.constant import TaskState
from formica.models.models import DeviceRunModel
from formica.models.models import DeviceSetModel
from formica.models.models import FlowRunModel
from formica.models.models import TaskModel
from formica.node.flow import Flow
from formica.settings import app_config
from formica.utils.session import NEW_SESSION
from formica.utils.session import provide_session
from sqlmodel.ext.asyncio.session import AsyncSession

logger = logging.getLogger(__name__)


def _create_tasks_for_device_run(
    device_run_model: DeviceRunModel, session: AsyncSession
) -> None:
    """Tạo các TaskModel từ thông tin của flow và device_run truyền vào, add vào session luôn"""
    flow = Flow(
        flow_version=device_run_model.flow_run.flow_version,
        arguments=device_run_model.flow_run.args,
    )
    for node_id, op in flow.node_dict.items():
        new_task = TaskModel(
            device_run_db_id=device_run_model.device_run_db_id,
            node_id=node_id,
            state=TaskState.WAIT_FOR_EXECUTING,
        )
        session.add(new_task)


class Scheduler:
    def __init__(self, executor: Executor, multi_processing: bool = True):
        self.executor = executor
        self.multi_processing = multi_processing

    @provide_session
    async def run(
        self, max_iteration: Optional[int] = None, session: AsyncSession = NEW_SESSION
    ):
        if isinstance(self.executor, LocalExecutor):
            logger.debug("executor is local")
        iteration = 0
        while True:
            if max_iteration is not None and iteration >= max_iteration:
                break

            # print("db url is", engine.url)
            flowruns_to_schedule = await FlowRunModel.get_flowrun_to_schedule(session)
            # Expunge these out of the session in order to add the processed one back
            # session.expunge_all()

            processes_do_any_work = []
            for flowrun_model in flowruns_to_schedule:
                temp = await self._schedule_flowrun(flowrun_model, session)
                processes_do_any_work.append(temp)

            # print("Try run executor...", len(self.executor._work_queue))
            await self.executor.run(session)

            iteration += 1
            if not any(processes_do_any_work):
                # logger.info(
                #     f"Didn't schedule anything this time, sleep for 2 seconds..."
                # )
                sleep(app_config.getint("scheduler", "POLLING_INTERVAL"))

        # TODO: Thông báo lý do gì mà vòng lặp schedule dừng
        logger.info(f"Đã lặp đủ số lần lặp schedule: {max_iteration}")
        logger.info("Dừng scheduler...")

    @provide_session
    async def _schedule_flowrun(
        self, flowrun_model: FlowRunModel, session: AsyncSession = NEW_SESSION
    ) -> bool:
        """
        :param flowrun_model: flow_run to process
        :param session: DB session
        :return: Is anything scheduled in this loop?
        """
        if (
            flowrun_model.state == FlowRunState.SUBMITTED
        ):  # Flow submitted, not scheduled
            return await self._schedule_submitted_flowrun(flowrun_model, session)
        else:
            return await self._schedule_running_flowrun(flowrun_model, session)

    async def _schedule_submitted_flowrun(
        self, flowrun_model: FlowRunModel, session: AsyncSession = NEW_SESSION
    ) -> bool:
        # Find the device set and loop through all devices in that set
        device_set = await DeviceSetModel.get_by_key(flowrun_model.device_set_id)
        for device in device_set.devices:
            # create DeviceRunModel
            device_run_model = DeviceRunModel(
                flow_id=flowrun_model.flow_id,
                version=flowrun_model.version,
                device_id=device.device_id,
                flow_run_id=flowrun_model.flow_run_id,
                device_run_id=f"{flowrun_model.run_type}__{time.time()}",
                logical_start_time=flowrun_model.start_time,
            )

            session.add(device_run_model)
            flowrun_model.set_state(FlowRunState.RUNNING)

            # Commit to get device_run_db_id
            await session.commit()
            await session.refresh(device_run_model)

            _create_tasks_for_device_run(device_run_model, session)
            await session.commit()
            # print("Enqueue:", device_run_model.device_run_id)
            self.executor.enqueue_device_run(device_run_model)
            print("size of queue", len(self.executor._work_queue))
            return True

        await session.commit()
        return False

    async def _schedule_running_flowrun(
        self, flowrun_model: FlowRunModel, session: AsyncSession = NEW_SESSION
    ) -> bool:
        latest_device_run = await DeviceRunModel.get_latest_device_run_of_flowrun(
            flowrun_model, session
        )
        if latest_device_run is None:
            # TODO: Handle invalid state - this FlowRun is RUNNING but has no DeviceRun
            raise Exception(
                f"FlowRun: {flowrun_model.flow_id}.{flowrun_model.version}.{flowrun_model.flow_run_id} is RUNNING but has no deviceRun"
            )
        if latest_device_run.state == DeviceRunState.SUCCESS:
            flowrun_model.state = FlowRunState.FINISHED
            await session.commit()
            return True
        elif latest_device_run.state == DeviceRunState.FAILED:
            flowrun_model.state = FlowRunState.FINISHED
            await session.commit()
            return True

        return False


async def _scheduler():
    scheduler_ = Scheduler(LocalExecutor())
    await scheduler_.run()


def main():
    asyncio.run(_scheduler())


if __name__ == "__main__":
    main()



from __future__ import annotations

import configparser
import json
import logging.config
import os
import pathlib
from importlib.resources import files

from sqlalchemy.ext.asyncio import create_async_engine

# print("Set start method")
# multiprocessing.set_start_method("spawn")

_DEFAULT_CONFIG = {
    "webserver": {"HOST": "127.0.0.1", "PORT": 8000},
    "scheduler": {
        "POLLING_INTERVAL": 1,
    },
    "executor": {"TYPE": "local"},
    "database": {"CONNECTION_POOL_SIZE": 1},
    "connection": {
        "DEFAULT_IDLE_TIMEOUT": 3000,
        "DEFAULT_CONNECT_TIMEOUT": 30,
        "DEFAULT_COMMAND_TIMEOUT": 30,
        "CONNECTION_MAX_RETRIES": 2,
        "CONNECTION_RETRY_SLEEP": 1,
    },
    "security": {
        "SECRET_KEY": "SECRET",
        "TOKEN_LIFETIME": 3600,  # 1 hour
    },
}

# ------------ PATHS -----------------------------------------------------------------------------------

_DEFAULT_FORMICA_HOME = pathlib.Path.home() / "formica"
FORMICA_HOME = (
    pathlib.Path(os.getenv("FORMICA_HOME"))
    if os.getenv("FORMICA_HOME") is not None
    else _DEFAULT_FORMICA_HOME
)

_CONFIG = FORMICA_HOME / "formica.ini"
_LOGGING_CONFIG = FORMICA_HOME / "logging_config.json"
_TEST_CONFIG = "src/formica/config_templates/test.ini"
_LOGS = FORMICA_HOME / "logs"
_SCHEDULER_LOGS = _LOGS / "scheduler"
_FLOW_LOGS = _LOGS / "flow"

# ------------ CONNECTIONS ------------------------------------------------------------------------------

_SQLITE_DB_TEST_FILE = "test_formica.db"
_SQLITE_DB_FILE = "formica.db"
_SQLITE_DEFAULT_CONNECTION_URL = f"sqlite+aiosqlite:///{FORMICA_HOME / _SQLITE_DB_FILE}"
_SQLITE_TEST_CONNECTION_URL = (
    f"sqlite+aiosqlite:///{FORMICA_HOME / _SQLITE_DB_TEST_FILE}"
)

# ------------- OBJECTS ---------------------------------------------------------------------------------

app_config = configparser.ConfigParser()
app_config.read_dict(_DEFAULT_CONFIG)
engine = None
_logger = logging.getLogger(__name__)
_init_already = False

# -------------------------------------------------------------------------------------------------------


def init() -> None:
    """
    Initialize the folders at `FORMICA_HOME`.

    This function should be called before running any components of Formica to set up necessary resources.

    It does not initialize the database (due to circular import), so the caller must call `create_all` of SQLModel afterwards.

    :return: None
    """
    global _init_already
    global FORMICA_HOME

    if not _init_already:
        print("Init already:", _init_already)
        formica_home_env = os.getenv("FORMICA_HOME", None)
        if formica_home_env is None:
            print(
                f"Environment variable FORMICA_HOME not found, default to {FORMICA_HOME}"
            )
        else:
            FORMICA_HOME = formica_home_env

        _create_formica_home()
        _init_config()
        _init_engine()

        print("Setting init_already to False...")
        _init_already = True


def _create_formica_home():
    if _formica_home_existed():
        return

    # Create logs folders
    FORMICA_HOME.mkdir(parents=True)
    _SCHEDULER_LOGS.mkdir(parents=True, exist_ok=True)
    _FLOW_LOGS.mkdir(parents=True, exist_ok=True)

    # Create a config file from template if it does not exist
    if not _CONFIG.exists():
        default_config_data = (
            files("formica.config_templates")
            .joinpath("default_formica.ini")
            .read_bytes()
        )
        # Write it to destination
        _CONFIG.write_bytes(default_config_data)


def _formica_home_existed():
    return FORMICA_HOME.exists() and FORMICA_HOME.is_dir()


def _init_config():
    _setup_logging()
    global app_config
    app_config.read(_CONFIG, encoding="utf-8")

    if os.getenv("TESTING", "") == "true":
        # Read the test config and override the values in the main config if they have the same key
        app_config.read(_TEST_CONFIG)


def _setup_logging():
    _logger.info("Setting up logging")
    if _LOGGING_CONFIG.exists():
        logging_config = json.loads(_LOGGING_CONFIG.read_text(encoding="utf-8"))
        logging.config.dictConfig(config=logging_config)
        return

    # If logging config file does not exist, create a default one
    default_config_data = (
        files("formica.config_templates")
        .joinpath("logging_config.json")
        .read_text(encoding="utf-8")
    )
    logging_config = json.loads(default_config_data)

    # Set log file path
    logging_config["handlers"]["file"]["filename"] = str(
        _SCHEDULER_LOGS / "formica.log"
    )
    logging.config.dictConfig(config=logging_config)
    _LOGGING_CONFIG.write_text(json.dumps(logging_config, indent=2), encoding="utf-8")

    _logger.info("Done setting up logging")


def _init_engine():
    global engine
    connection_url = app_config.get(
        "database", "DATABASE_URL", fallback=_SQLITE_DEFAULT_CONNECTION_URL
    )
    print("connection_url:", connection_url)

    # Create the engine based on the current environment,
    # if it is a TESTING environment, create an in-memory sqlite engine
    if os.environ.get("TESTING", "") == "true":
        engine = create_async_engine(_SQLITE_TEST_CONNECTION_URL, echo=False)
    else:
        engine = create_async_engine(
            connection_url,
            pool_size=app_config.getint("database", "CONNECTION_POOL_SIZE"),
            echo=False,
        )


init()



import asyncio
import logging
import re
import time
from abc import ABC
from abc import abstractmethod
from asyncio import InvalidStateError
from typing import Optional
from typing import override

import asyncssh
import telnetlib3
from formica.models.constant import ConnectionType
from formica.models.constant import DEFAULT_CONNECT_TIMEOUT
from formica.models.constant import DEFAULT_IDLE_TIMEOUT
from formica.models.constant import DeviceType
from formica.models.models import CredentialModel
from formica.models.models import DeviceModel
from telnetlib3 import TelnetReader
from telnetlib3 import TelnetWriter


logger = logging.getLogger(__file__)


def _expect(text: str, patterns: list[str]) -> tuple[bool, str]:
    """True if any of the patterns is found in the text"""
    for pattern in patterns:
        if re.search(pattern, text):
            return True, pattern

    return False, ""


class BaseConnection(ABC):
    def __init__(
        self,
        host: str,
        idle_timeout: float = DEFAULT_IDLE_TIMEOUT,
        prompt: Optional[str] = "",
    ):
        self.host: str = host
        self._idle_timeout = idle_timeout
        self.running_command: Optional[str] = None
        self.prompt: str = prompt
        self._stdout_future: Optional[asyncio.Future[str]] = None
        self._last_active: float = time.time()
        self.current_command_timeout: Optional[float] = None

    def idle_timed_out(self) -> bool:
        """Is this connection idle for too long?"""
        if self.running_command is not None:
            return False

        return time.time() - self._last_active > self._idle_timeout

    @abstractmethod
    async def run_command(
        self, command: str, expect_prompt: Optional[str], timeout: float
    ) -> str:
        pass

    @abstractmethod
    async def close(self) -> None:
        pass


class SSHConnection(BaseConnection):
    """Don't create this class directly using the constructor, use ConnectionFactory instead"""

    default_expect_prompts = [r">", r"#", r"assword"]

    def __init__(
        self,
        connection: asyncssh.SSHClientConnection,
        process: Optional[asyncssh.SSHClientProcess],
        device_type: DeviceType,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._connection = connection
        self._process = process
        self._device_type = device_type

    @classmethod
    async def expect_response(
        cls, process: asyncssh.SSHClientProcess, patterns, timeout=5
    ) -> Optional[str]:
        """Read the output for some time, return the output if any pattern is matched"""
        buffer = ""
        try:
            async with asyncio.timeout(timeout):
                while True:
                    chunk = await process.stdout.read(1024)
                    if len(chunk) != 0:
                        # print("chunk:", chunk.encode())
                        buffer += chunk
                        # print(f"Buffer: {buffer.encode()}")  # Debug log

                        for pattern in patterns:
                            if re.search(pattern, buffer):
                                print(f"Matched pattern: {pattern}")
                                return buffer

                        if "ENTER" in buffer:
                            buffer = buffer.rsplit("\n", 1)[0]

                            process.stdin.write("\n")
                            await process.stdin.drain()
                    else:
                        # No more data to read, break the loop
                        print("No more data to read from stdout")
                        break
        except asyncio.TimeoutError:
            pass
        return buffer

    @override
    async def run_command(
        self, command: str, expect_prompt: Optional[str], timeout: float
    ) -> str:
        """
        Schedule a command to run on this connection

        Args:
            command: The command to run on this connection
            expect_prompt: The expected prompt to wait for after command output
            timeout: Timeout for executing command

        Raises:
            InvalidStateError: When a command is already running
            ValueError: If no credentials found or if connection type not supported
        """
        if self.running_command is not None:
            raise InvalidStateError(
                f"A command is already running: {self.running_command}"
            )

        print(f"Run command: {command}")

        # Get Future for stdout
        if self._device_type == DeviceType.LINUX:
            self._stdout_future = asyncio.create_task(self._connection.run(command))
        elif self._device_type == DeviceType.GCOM:
            self._process.stdin.write(f"{command}\n")
            await self._process.stdin.drain()
            print(
                f"expect prompt is: {self.prompt if expect_prompt is None else expect_prompt}"
            )
            self._stdout_future = asyncio.create_task(
                SSHConnection.expect_response(
                    self._process,
                    [self.prompt if expect_prompt is None else expect_prompt],
                )
            )
        else:
            self._process.stdin.write(f"{command}\n")
            await self._process.stdin.drain()
            print(
                f"expect prompt is: {self.prompt if expect_prompt is None else expect_prompt}"
            )
            self._stdout_future = asyncio.create_task(
                SSHConnection.expect_response(
                    self._process,
                    [self.prompt if expect_prompt is None else expect_prompt],
                    timeout=5,
                )
            )

        self.running_command = command
        self._last_active = time.time()

        try:
            await self._stdout_future
            return self._get_output()
        except Exception as e:
            raise e
        except asyncio.TimeoutError as e:
            raise TimeoutError(
                f"Timeout when running command {command} on {self.host}: {e}"
            )

    def _get_output(self) -> str:
        """
        Get the result of the running command

        Returns:
            Tuple contains the command, and result of that command, or None, plus host at the end

        Raises:
            InvalidStateError: When no command is running
        """
        # Check if the task is done, else return False and empty string
        if self.running_command is None:
            raise InvalidStateError("No running command")

        if e := self._stdout_future.exception() is not None:
            raise ConnectionError(
                f"Exception when running command {self.running_command}: {e}"
            )

        if self._device_type == DeviceType.LINUX:
            result = self._stdout_future.result().stdout
            print("result is", result)
        else:
            result = self._stdout_future.result()
            logger.info(f"result running command {self.running_command} is: {result}")

        self._stdout_future = None
        self.running_command = None
        self.current_command_timeout = None

        return result

    @override
    async def close(self) -> None:
        self.running_command = None
        self._stdout_future = None
        self.current_command_timeout = None
        self._connection.close()
        await self._connection.wait_closed()


class TelnetConnection(BaseConnection):
    default_expect_usernames = [r"sername", r"ogin", r"ser name"]
    default_expect_passwords = [r"assword"]
    default_expect_prompts = [r">", r"#"]
    default_expect_errors = [r"error", r"Error"]
    """Don't create this class directly using the constructor, use ConnectionFactory instead"""

    def __init__(
        self,
        reader: TelnetReader,
        writer: TelnetWriter,
        device_type: DeviceType,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._reader: TelnetReader = reader
        self._writer: TelnetWriter = writer
        self._device_type = device_type

    # @classmethod
    # async def open_connection(cls, ip, port) -> tuple[TelnetReader, TelnetWriter]:
    #     error_string = ""
    #     for try_number in range(TELNET_MAX_RETRIES):
    #         try:
    #             print(
    #                 f"Try to connect to {ip} using telnet, port: {port}"
    #             )
    #             reader, writer = await telnetlib3.open_connection(ip, port, encoding="latin-1")
    #             return reader, writer
    #         except Exception as e:
    #             error_string += f"\n- {e}"
    #             print(f"Error connecting to {ip} using telnet (try number {try_number}): {str(e)}")
    #
    #     raise ConnectionError(f"Failed to connect to {ip} using telnet: {error_string}")

    @classmethod
    async def expect_response(
        cls, reader: TelnetReader, patterns, timeout=10
    ) -> tuple[str, str]:
        """Wait for any pattern from the list, checking received data in chunks."""
        buffer = ""
        try:
            while True:
                chunk = await asyncio.wait_for(
                    reader.read(1024), timeout=timeout
                )  # Read in chunks
                # print(f"Type of chunk: {type(chunk)}")
                # print(f"chunk: {chunk.encode()}")
                buffer += chunk
                # print(f"Type of buffer: {type(buffer)}")
                # print(f"Buffer: {buffer.encode()}")  # Debug log

                for pattern in patterns:
                    if re.search(pattern, buffer):
                        print(f"Matched pattern: {pattern}")
                        return buffer, pattern
        except asyncio.TimeoutError:
            print("Timeout while waiting for response")
            return buffer, ""

    @classmethod
    async def expect_response2(
        cls, reader: TelnetReader, patterns: list[str]
    ) -> Optional[str]:
        """Wait for a specific pattern in the response"""
        found = False
        while not found:
            response = await reader.readuntil(b"\n")  # Read a line
            print(f"response to expect {patterns} from:")
            print(response)

            decoded_response = response.decode("utf-8").strip()

            for pattern in patterns:
                if re.search(pattern, decoded_response):
                    return decoded_response
        return None

    @classmethod
    async def send_command(cls, writer: TelnetWriter, command: str) -> None:
        writer.write(command + "\n")
        await writer.drain()

    async def run_command(
        self, command: str, expect_prompt: Optional[str], timeout: float
    ) -> str:
        """
        Schedule a command to run on this connection

        Args:
            command: The command to run on this connection
            expect_prompt: The expected prompt to wait for after command output
            timeout: Timeout for executing command

        Raises:
            InvalidStateError: When a command is already running
            ValueError: If no credentials found or if connection type not supported
        """
        if self.running_command is not None:
            raise InvalidStateError(
                f"A command is already running: {self.running_command}"
            )

        print(f"Run command: {command}")

        self._stdout_future = asyncio.create_task(
            asyncio.wait_for(
                self._schedule_run_command(command, expect_prompt), timeout=timeout
            )
        )
        self.running_command = command
        self.current_command_timeout = timeout
        self._last_active = time.time()

        try:
            await self._stdout_future
            return self._get_output()
        except Exception as e:
            raise e

    async def _schedule_run_command(
        self, command: str, expect_prompt: Optional[str]
    ) -> str:
        await TelnetConnection.send_command(self._writer, command)
        print(
            f"expect prompt is: {self.prompt if expect_prompt is None else expect_prompt}"
        )
        output = await self._reader.readuntil(
            self.prompt.encode("latin-1")
            if expect_prompt is None
            else expect_prompt.encode("latin-1")
        )
        output = output.decode("latin-1")  # Decode
        _, output = output.split("\n", 1)
        return output

    def _get_output(self) -> str:
        """
        Get the result of the running command

        Returns:
            Tuple contains the command, and result of that command, or None, plus host at the end

        Raises:
            InvalidStateError: When no command is running
        """
        # Check if the task is done, else return False and empty string
        if self.running_command is None:
            raise InvalidStateError("No running command")
        if e := self._stdout_future.exception() is not None:
            raise ConnectionError(
                f"DEBUG: Exception when running command {self.running_command}: {e}"
            )

        result = self._stdout_future.result()
        print("result of command:", result)

        self._stdout_future = None
        self.running_command = None
        self.current_command_timeout = None

        return result

    @override
    async def close(self) -> None:
        self.running_command = None
        self._stdout_future = None
        self.current_command_timeout = None
        self._writer.close()
        self._reader.feed_eof()


class ConnectionFactory:
    @staticmethod
    async def new_connection(
        device: DeviceModel,
        credential: CredentialModel,
        client_keys: Optional[list[str]],
        connect_timeout: int,
        idle_timeout: float = DEFAULT_IDLE_TIMEOUT,
    ) -> BaseConnection:
        connection_type = credential.connection_type
        try:
            if connection_type == ConnectionType.SSH:
                print("SSHing...")
                return await ConnectionFactory._new_ssh_connection(
                    device, credential, client_keys, connect_timeout, idle_timeout
                )
            elif connection_type == ConnectionType.TELNET:
                return await ConnectionFactory._new_telnet_connection(
                    device, credential, connect_timeout, idle_timeout
                )
            else:
                raise ConnectionError(f"Unknown connection type: {connection_type}")
        except (TimeoutError, ConnectionError, Exception) as e:
            raise e

    @staticmethod
    async def _new_ssh_connection(
        device: DeviceModel,
        credential: CredentialModel,
        client_keys: Optional[list[str]],
        connect_timeout: int,
        idle_timeout: float = DEFAULT_IDLE_TIMEOUT,
    ) -> BaseConnection:
        """
        Schedule a command to run on this connection

        Returns:
            The connection object (if connect success)

        Raises:
            ConnectionError: If connect fails
        """
        host = device.ip
        password = None if credential.password == "" else credential.password
        device_type = device.device_type
        # device_type = DeviceType.get_device_type(device_model.code)
        print(f"DeviceModel type: {device_type}")

        try:
            print("Connecting to device using SSH...")
            if client_keys is None:
                conn = await asyncssh.connect(
                    host=device.ip,
                    port=credential.port,
                    username=credential.username,
                    password=password,
                    known_hosts=None,
                    connect_timeout=connect_timeout,
                    encryption_algs="*",
                    kex_algs="*",
                )
            else:
                conn = await asyncssh.connect(
                    host=device.ip,
                    port=credential.port,
                    username=credential.username,
                    password=password,
                    client_keys=client_keys,
                    known_hosts=None,
                    connect_timeout=connect_timeout,
                    encryption_algs="*",
                    kex_algs="*",
                )

            if device_type == DeviceType.H3C:
                process = await conn.create_process(
                    term_type="ansi", encoding="latin-1"
                )
            else:
                process = await conn.create_process(
                    term_type="ansi", encoding="latin-1"
                )
            prompt = ">"

            # Prepare the connection
            if device_type != DeviceType.LINUX:
                print("wait for initial prompt")
                initial_response = await SSHConnection.expect_response(
                    process, SSHConnection.default_expect_prompts
                )
                if not initial_response:
                    raise ConnectionError(
                        f"Cannot connect to {host}, expected patterns: {SSHConnection.default_expect_prompts}, but got no response"
                    )

                print("init 1:", initial_response.encode())
                if "assword" in initial_response:
                    # Have to enter the password to login
                    process.stdin.write(f"{password}\n")
                    await process.stdin.drain()
                    initial_response = await SSHConnection.expect_response(
                        process, SSHConnection.default_expect_prompts
                    )
                    print("init 2:", initial_response.encode())
                    if prompt not in initial_response:
                        raise ConnectionError(
                            f"Cannot connect, may be the credential is incorrect. Server message: {initial_response}"
                        )

                logger.info("Get the last line as prompt")
                prompt = initial_response.split("\n")[-1]
                print(f"initial prompt: {prompt}")
                # if device_model.code not in prompt:
                #     raise ConnectionError(
                #         f"Wrong type of device, expected {device_model.code} but got {prompt}"
                #     )

                await ConnectionFactory._disable_paging(device_type, process, prompt)

            print("Done, returning connection...")

            return SSHConnection(
                connection=conn,
                process=process,
                device_type=device_type,
                idle_timeout=idle_timeout,
                host=host,
                prompt=prompt,
            )
        except TimeoutError as e:
            raise e
        except Exception as e:
            raise ConnectionError(f"Failed to connect to {host}: {str(e)}")

    @classmethod
    async def _disable_paging(
        cls, device_type: DeviceType, process: asyncssh.SSHClientProcess, prompt: str
    ) -> None:
        """
        Disable paging for the device, so that we can get the full output of the command

        Raises:
            TimeoutError: If the command to disable paging times out
            ConnectionError: If the command to disable paging fails
        """
        if device_type == DeviceType.HUAWEI:
            process.stdin.write("screen-length 0 temporary\n")
            output = await SSHConnection.expect_response(process, [prompt], timeout=15)

            # Run the scroll command instead
            if "error" in output:
                process.stdin.write("scroll\n")
                output = await SSHConnection.expect_response(
                    process, [prompt], timeout=15
                )

            # Error
            if output == "":
                raise TimeoutError(
                    "Failed to disable paging using command: screen-length 0 temporary, scroll, timeout while waiting for prompt"
                )
            if "error" in output:
                raise ConnectionError(
                    f"Failed to disable paging using command \n- screen-length 0 temporary\n- scroll 512\nOutput: {output}"
                )

            logger.debug(f"disable paging output: \n{output}")
            print(f"disable paging output: \n{output}")
        elif device_type == DeviceType.H3C:
            print("Disable paging for H3C: screen-length disable")
            process.stdin.write("screen-length disable\n")
            output = await SSHConnection.expect_response(process, [prompt], timeout=15)

            # Error
            if output == "":
                raise TimeoutError(
                    "Failed to disable paging using command: screen-length disable, timeout while waiting for prompt"
                )
            if "error" in output:
                raise ConnectionError(
                    f"Failed to disable paging using command \n- screen-length disable\nOutput: {output}"
                )

            logger.debug(f"disable paging output: \n{output}")
            print(f"disable paging output: \n{output}")
        else:
            return

    @classmethod
    async def _new_telnet_connection(
        cls,
        device: DeviceModel,
        credential: CredentialModel,
        connect_timeout: int = DEFAULT_CONNECT_TIMEOUT,
        idle_timeout: float = DEFAULT_IDLE_TIMEOUT,
    ) -> BaseConnection:
        """
        Try to connect and authenticate using telnet

        Returns:
            The connection object (if connect success)

        Raises:
            TimeoutError: If connect timeout for some reason
        """
        reader: TelnetReader
        writer: TelnetWriter
        print(f"Getting device model with id: {credential.device_model_id}")
        # device_model = await get_device_model(credential.device_model_id)
        device_model = device.device_model
        print(f"Got device model: {device_model}")
        device_type = DeviceType.get_device_type(device_model.code)
        print(f"DeviceModel type: {device_type}")
        try:
            print(
                f"Try to connect to {device.ip} using telnet, port: {credential.port}"
            )
            reader, writer = await asyncio.wait_for(
                telnetlib3.open_connection(device.ip, credential.port),
                timeout=connect_timeout,
            )
        except asyncio.TimeoutError:
            raise TimeoutError(
                f"Timeout error (connect took longer than {connect_timeout} seconds)"
            )

        # await asyncio.sleep(1)  # Wait for negotiation?

        if credential.username is not None and credential.username:
            print("Doing authentication")
            try:
                await cls._authenticate_telnet(credential, reader, writer)
            except TimeoutError as e:
                raise e

        expect_prompts = (
            TelnetConnection.default_expect_prompts
            if "expect_prompts" not in credential.extra
            else [re.escape(pattern) for pattern in credential.extra["expect_prompts"]]
        )

        expect_errors = (
            TelnetConnection.default_expect_errors
            if "expect_prompts" not in credential.extra
            else [re.escape(pattern) for pattern in credential.extra["expect_errors"]]
        )

        # Get the prompt phrase
        logger.info("wait for initial prompt")
        response, pattern = await TelnetConnection.expect_response(
            reader, expect_prompts + expect_errors
        )
        match, pattern = _expect(response, expect_prompts + expect_errors)

        if not match:
            raise ConnectionError(
                f"Cannot connect to {device.ip}, expected patterns: {expect_prompts + expect_errors}, but got: {response.encode()}"
            )

        if pattern in TelnetConnection.default_expect_errors:
            raise ConnectionError(
                f"Failed to connect to {device.ip}: {response.encode()}"
            )

        prompt = response.split("\n")[-1]
        print(f"Prompt is {prompt}")

        return TelnetConnection(
            host=device.ip,
            reader=reader,
            writer=writer,
            device_type=device_type,
            prompt=prompt,
            idle_timeout=idle_timeout,
        )

    @classmethod
    async def _authenticate_telnet(
        cls,
        credential: CredentialModel,
        reader,
        writer,
        timeout=DEFAULT_CONNECT_TIMEOUT,
    ):
        """
        Authenticate

        Returns:
            The connection object (if connect success)

        Raises:
            TimeoutError: If connect timeout (cannot find the expected pattern)
        """
        expect_usernames = (
            TelnetConnection.default_expect_usernames
            if "expect_usernames" not in credential.extra
            else [
                re.escape(pattern) for pattern in credential.extra["expect_usernames"]
            ]
        )
        expect_passwords = (
            TelnetConnection.default_expect_passwords
            if "expect_passwords" not in credential.extra
            else [
                re.escape(pattern) for pattern in credential.extra["expect_passwords"]
            ]
        )
        # Authenticate:
        try:
            response_username_prompt, _ = await TelnetConnection.expect_response(
                reader, expect_usernames, timeout=timeout
            )
            # if not _expect(response_username_prompt, expect_usernames):

            print("response (expect username):", response_username_prompt.encode())
            print(f"login using username: {credential.username}")
            writer.write(credential.username + "\n")
            await writer.drain()

            response_password_prompt, _ = await TelnetConnection.expect_response(
                reader, expect_passwords, timeout=timeout
            )
            print("response (expect password):", response_password_prompt.encode())
            print(f"login using password: {credential.password}")
            writer.write(credential.password + "\n")
            await writer.drain()
        except TimeoutError as e:
            print(f"Timeout error when authenticate telnet: {e}")
            raise



from typing import Annotated

from fastapi import APIRouter
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Path
from fastapi import Query
from formica.models.constant import UserRole
from formica.models.models import DeviceModel
from formica.models.models import DeviceSetModel
from formica.models.models import User
from formica.models.request_models import CreateDeviceSet
from formica.models.request_models import DeviceSetFilters
from formica.models.request_models import DeviceSetResponse
from formica.models.request_models import ResponseDeviceSet
from formica.models.request_models import UpdateDevicesOfDeviceSet
from formica.utils.session import get_session
from formica.web.users import current_active_user
from sqlmodel.ext.asyncio.session import AsyncSession
from starlette.status import HTTP_200_OK
from starlette.status import HTTP_201_CREATED
from starlette.status import HTTP_204_NO_CONTENT
from starlette.status import HTTP_400_BAD_REQUEST
from starlette.status import HTTP_403_FORBIDDEN

SessionDep = Annotated[AsyncSession, Depends(get_session)]
device_set_router = APIRouter()


@device_set_router.get(
    path="/{device_set_id}", description="Get a device set", status_code=HTTP_200_OK
)
async def get_device_set(
    device_set_id: str, session: SessionDep, user: User = Depends(current_active_user)
) -> ResponseDeviceSet:
    device_set = await DeviceSetModel.get_by_key(device_set_id, session)

    if device_set is None:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail="Device set not found"
        )

    if device_set.group_id not in (await user.get_group_ids(session)):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Can't get device set in group {device_set.group_id} because you are not a member of this group",
        )

    return ResponseDeviceSet(
        device_set_id=device_set.device_set_id,
        description=device_set.description,
        devices=[device.device_id for device in device_set.devices],
        group_id=device_set.group_id,
    )


@device_set_router.get(path="", description="Get device sets", status_code=HTTP_200_OK)
async def get_device_sets(
    filters: Annotated[DeviceSetFilters, Query()],
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> list[ResponseDeviceSet]:
    db_device_sets = await DeviceSetModel.filter(user, filters, session)
    return [
        ResponseDeviceSet(
            device_set_id=device_set.device_set_id,
            description=device_set.description,
            devices=[device.device_id for device in device_set.devices],
            group_id=device_set.group_id,
        )
        for device_set in db_device_sets
    ]


@device_set_router.post(
    path="", description="Create a device set", status_code=HTTP_201_CREATED
)
async def add_device_set(
    new_device_set: CreateDeviceSet,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> CreateDeviceSet:
    if not (user.is_superuser or user.role == UserRole.ADMIN):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to create a device set",
        )

    if new_device_set.group_id not in (await user.get_group_ids(session)):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail=f"Can't create device set in group {new_device_set.group_id} because you are not a member of this group",
        )

    db_device_set = DeviceSetModel(
        device_set_id=new_device_set.device_set_id,
        description=new_device_set.description,
        group_id=new_device_set.group_id,
    )
    not_found_devices = []

    for device_id in new_device_set.devices:
        device = await DeviceModel.get_by_key(new_device_set.group_id, device_id)
        if device is None:
            not_found_devices.append(device_id)
        else:
            db_device_set.devices.append(device)

    if not_found_devices:
        raise HTTPException(
            status_code=400,
            detail=f"Devices with these ids not found: {'\n'.join(not_found_devices)}",
        )

    session.add(db_device_set)
    await session.commit()
    await session.refresh(db_device_set)

    return new_device_set


# @device_set_router.patch(
#     path="",
#     description="Thêm một hoặc nhiều device vào một device set có sẵn",
# )
# async def update_device_set(
#     update_device_set: UpdateDeviceSet, session: SessionDep
# ) -> DeviceSetModel:
#     db_device_set = await DeviceSetModel.get_device_set_by_id(
#         update_device_set.device_set_id
#     )
#
#     for device_id in update_device_set.device_id_list:
#         device = DeviceModel.get_device_by_device_id(device_id)
#         if device is None:
#             raise HTTPException(
#                 status_code=400, detail=f"Device with this id not found: {device_id}"
#             )
#
#         db_device_set.devices.append(device)
#
#     session.add(db_device_set)
#     await session.commit()
#     await session.refresh(db_device_set)
#
#     return db_device_set


@device_set_router.delete(
    path="/{device_set_id}",
    description="Delete a device set",
    status_code=HTTP_204_NO_CONTENT,
)
async def remove_device_set(
    device_set_id: str, session: SessionDep, user: User = Depends(current_active_user)
) -> None:
    if not (user.is_superuser or user.role == UserRole.ADMIN):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to delete a device",
        )

    device_set = await DeviceSetModel.get_by_key(device_set_id, session)

    if device_set.group_id not in (await user.get_group_ids(session)):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail=f"Your groups do not have this device set: {device_set_id}",
        )

    if device_set is not None:
        await session.delete(device_set)
        await session.commit()


@device_set_router.post(
    path="/{device_set_id}/devices",
    description="Update devices of a device set",
    status_code=HTTP_200_OK,
)
async def update_devices(
    device_set_id: Annotated[str, Path(title="The target group id to update")],
    update_devices_request: UpdateDevicesOfDeviceSet,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> DeviceSetResponse:
    if not (user.is_superuser or user.role == UserRole.ADMIN):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to update devices of a device set",
        )

    db_device_set: DeviceSetModel = await DeviceSetModel.get_by_key(
        device_set_id, session
    )
    if db_device_set is None:
        raise HTTPException(status_code=400, detail="Device set not existed")

    if db_device_set.group_id not in (await user.get_group_ids(session)):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail=f"Can't update devices in device set {device_set_id} because you are not a member of the group it belongs to",
        )

    devices = await DeviceModel.get_devices_by_id(
        update_devices_request.devices, session
    )
    db_device_set.devices = devices
    session.add(db_device_set)
    await session.commit()

    response = DeviceSetResponse(
        device_set_id=db_device_set.group_id,
        description=db_device_set.description,
        devices=[device.device_id for device in db_device_set.devices],
    )

    return response



import logging
from pathlib import Path

from sqlmodel.ext.asyncio.session import AsyncSession

from formica.execution.context import Context
from formica.execution.device_run import DeviceRun
from formica.models.models import DeviceRunModel
from formica.node.flow import Flow
from formica.settings import FORMICA_HOME
from formica.utils.session import NEW_SESSION
from formica.utils.session import provide_session

logger = logging.getLogger(__name__)


def _prepare_log_file_path(device_run: DeviceRunModel) -> Path:
    log_file_dir = (
        FORMICA_HOME
        / "logs"
        / f"flow_id={device_run.flow_id}"
        / f"version={device_run.version}"
        / f"flow_run={device_run.flow_run_id}"
    )
    log_file_dir.mkdir(parents=True, exist_ok=True)

    return log_file_dir / f"{device_run.device.device_id}.log"


@provide_session
async def execute_device_run(retry_run_db_id: int, session: AsyncSession = NEW_SESSION):
    print("Executing retry run...")
    print("another one", retry_run_db_id, session)
    try:
        device_run_model = await session.get(DeviceRunModel, retry_run_db_id)
        print(
            "Got device run model: ",
            device_run_model.flow_run_id,
            device_run_model.device_id,
        )

        flow = Flow(
            flow_version=device_run_model.flow_run.flow_version,
            arguments=device_run_model.flow_run.args,
        )
        print("Creating context...")
        log_file_path = _prepare_log_file_path(device_run_model)
        context = Context(flow, session, log_file_path, device_run_model.device)
        print("Creating context done")
        # TODO: Nếu None thì sao?
        if device_run_model is not None:
            device_run = DeviceRun(device_run_model, context)
            await device_run.run()
    except Exception as e:
        print("Error: ", e)



from typing import Annotated

from fastapi import APIRouter
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Query
from formica.models.constant import UserRole
from formica.models.models import CredentialModel
from formica.models.models import DeviceModel
from formica.models.models import User
from formica.models.request_models import DeviceFilters
from formica.utils.session import get_session
from formica.web.users import current_active_user
from sqlmodel.ext.asyncio.session import AsyncSession
from starlette.status import HTTP_201_CREATED
from starlette.status import HTTP_400_BAD_REQUEST
from starlette.status import HTTP_403_FORBIDDEN

SessionDep = Annotated[AsyncSession, Depends(get_session)]
credential_router = APIRouter()


@credential_router.post(
    path="",
    description="Create a credential to connect to a device",
    status_code=HTTP_201_CREATED,
)
async def add_credential(
    new_credential: CredentialModel,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> CredentialModel:
    # if not ConnectionType.is_valid(new_credential.connection_type):
    #     raise HTTPException(
    #         status_code=HTTP_400_BAD_REQUEST,
    #         detail=f"Connection type is not valid: {new_credential.connection_type}",
    #     )

    if not (user.is_superuser or user.role == UserRole.ADMIN):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to create a credential",
        )

    device = await DeviceModel.get_by_key(
        new_credential.group_id, new_credential.device_id, session
    )

    if device is None:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail="Device not found")

    if device.group_id not in (await user.get_group_ids(session)):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="Can't create credential of device in group that you are not a member of",
        )

    session.add(new_credential)
    await session.commit()
    await session.refresh(new_credential)

    return new_credential


@credential_router.get(path="", description="Get credentials")
async def get_credentials(
    session: SessionDep,
    filters: Annotated[DeviceFilters, Query()],
    user: User = Depends(current_active_user),
) -> list[CredentialModel]:
    return await CredentialModel.filter(user, filters, session)


@credential_router.delete(path="/{credential_id}", description="Delete a credential")
async def remove_credential(
    credential_id: str, session: SessionDep, user: User = Depends(current_active_user)
) -> None:
    if not (user.is_superuser or user.role == UserRole.ADMIN):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to delete a credential",
        )

    credential = await CredentialModel.get_by_key(credential_id, session)

    if credential.device.group_id not in (await user.get_group_ids(session)):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="Can't delete credential of device in group that you are not a member of",
        )

    if credential is not None:
        await session.delete(credential)
        await session.commit()



from enum import Enum
from typing import Optional

from formica.settings import app_config


class ValidateEnum(Enum):
    @classmethod
    def is_valid(cls, enum_str: str) -> bool:
        print(cls.__members__.values())
        return enum_str in cls.__members__.values()


class FlowRunState(str, ValidateEnum):
    SUBMITTED = "submitted"
    RUNNING = "running"
    FINISHED = "finished"


class FlowRunType(str, ValidateEnum):
    MANUAL = "manual"
    SCHEDULE = "schedule"


class DeviceRunState(str, ValidateEnum):
    QUEUED = "queued"
    RUNNING = "running"
    SUCCESS = "success"
    FAILED = "failed"


class TaskState(str, ValidateEnum):
    WAIT_FOR_EXECUTING = "wait_for_executing"
    SKIPPED = "skipped"
    RUNNING = "running"
    SUCCESS = "success"
    FAILED = "failed"


class FlowRetryMode(str, ValidateEnum):
    SKIP_SUCCESS = "skip_success"
    RETRY_ALL = "retry_all"


class TaskTriggerRule(str, ValidateEnum):
    """Điều kiện để toán tử được chạy"""

    SUCCESS = "success"  # Toán tử trước đó phải thành công
    DONE = "done"  # Toán tử trước đó chỉ cần chạy xong


class UserRole(str, ValidateEnum):
    ADMIN = "admin"
    OPERATOR = "operator"


FINISHED_TASK_STATES = (TaskState.SUCCESS, TaskState.FAILED, TaskState.SKIPPED)
INITIAL_FLOW_STATE = {
    "nodes": [
        {
            "id": "1",
            "type": "customNode",
            "position": {"x": 250, "y": 100},
            "data": {
                "label": "Start",
                "type": "trigger",
                "description": "Begins the workflow",
                "inputs": [],
                "outputs": [{"id": "out-1", "label": "Output"}],
                "config": {},
            },
        }
    ]
}


class ConnectionType(ValidateEnum):
    SSH = "ssh"
    TELNET = "telnet"
    HTTP = "http"
    HTTPS = "https"
    FTP = "ftp"
    SFTP = "sftp"


class DeviceType(ValidateEnum):
    LINUX = "linux"
    HUAWEI = "huawei"
    JUNIPER = "juniper"
    GCOM = "gcom"
    ZTE = "zte"
    H3C = "h3c"

    @classmethod
    def get_device_type(cls, code: str) -> Optional["DeviceType"]:
        if "HW" in code or "HA" in code:
            return DeviceType.HUAWEI
        elif "GC" in code:
            return DeviceType.GCOM
        elif "CH" in code:
            return DeviceType.H3C
        else:
            return None


class CatalogType(ValidateEnum):
    OLT = "olt"
    SWITCH = "switch"


DEFAULT_IDLE_TIMEOUT = app_config.getint("connection", "DEFAULT_IDLE_TIMEOUT")
DEFAULT_CONNECT_TIMEOUT = app_config.getint("connection", "DEFAULT_CONNECT_TIMEOUT")
DEFAULT_COMMAND_TIMEOUT = app_config.getint("connection", "DEFAULT_COMMAND_TIMEOUT")

CONNECTION_MAX_RETRIES = app_config.getint("connection", "CONNECTION_MAX_RETRIES")
CONNECTION_RETRY_SLEEP = app_config.getint("connection", "CONNECTION_RETRY_SLEEP")



import asyncio
import json
import logging
import multiprocessing

import aiohttp
from formica import settings
from formica.models.constant import ConnectionType
from formica.models.constant import DeviceType
from formica.models.constant import UserRole
from formica.models.models import CredentialModel
from formica.models.models import DeviceModel
from formica.models.models import DeviceSetModel
from formica.models.models import GroupModel
from formica.models.models import User
from formica.utils.session import NEW_SESSION
from formica.utils.session import provide_session
from formica.web.main import run_webserver
from sqlmodel import SQLModel
from sqlmodel.ext.asyncio.session import AsyncSession

logger = logging.getLogger(__name__)


async def start_api_server() -> multiprocessing.Process:
    process = multiprocessing.Process(target=run_webserver, args=[7999], daemon=False)
    print("Staring API server in as a daemon...")
    process.start()

    # Wait for the server to start?
    await asyncio.sleep(10)
    return process


def get_groups_data() -> dict:
    with open(
        "src/formica/utils/mock_data/mock_data.json", "r", encoding="utf-8"
    ) as file:
        return json.load(file)


def get_users_data() -> dict:
    with open(
        "src/formica/utils/mock_data/mock_users.json", "r", encoding="utf-8"
    ) as file:
        return json.load(file)


@provide_session
async def insert_users(users_mock_data: dict, session: AsyncSession = NEW_SESSION):
    async with aiohttp.ClientSession() as aiohttp_session:
        for user in users_mock_data["users"]:
            payload = {
                "email": user["email"],
                "name": user["name"],
                "password": user["password"],
            }
            async with aiohttp_session.post(
                "http://localhost:8000/api/auth/register", json=payload
            ) as response:
                if response.status == 201:
                    print(f"User {user['name']} created successfully.")
                else:
                    print(f"Failed to create user {user['name']}.")

    # Manually fix the is_active and is_superuser fields
    for user in users_mock_data["users"]:
        user_db: User = await User.get_user_by_email(user["email"], session=session)
        user_db.is_active = user["is_active"]
        user_db.is_superuser = user["is_superuser"]
        user_db.role = UserRole(user["role"])

    await session.commit()


@provide_session
async def insert_groups(groups_mock_data: dict, session: AsyncSession = NEW_SESSION):
    for group in groups_mock_data["groups"]:
        # Insert the groups
        group_db = GroupModel(
            group_id=group["group_id"], description=group["description"]
        )
        session.add(group_db)

        # Add the members for the groups
        group_db = await session.get(GroupModel, group["group_id"])
        members_to_add = await User.get_user_by_emails(
            group["members"], session=session
        )
        for member in members_to_add:
            group_db.members.append(member)

        # Insert the devices
        for device in group["devices"]:
            device_db = DeviceModel(
                device_id=device["device_id"],
                device_type=DeviceType(device["device_type"]),
                group_id=group["group_id"],
                ip=device["ip"],
            )
            session.add(device_db)

            # Insert the credentials of devices
            for credential in device["credentials"]:
                credential["connection_type"] = ConnectionType(
                    credential["connection_type"]
                )
                credential["group_id"] = group["group_id"]
                device_db.credentials.append(CredentialModel(**credential))

        # Create the device sets
        for device_set in group["device_sets"]:
            device_set_db = DeviceSetModel(
                device_set_id=device_set["device_set_id"],
                description=device_set["description"],
                group_id=group["group_id"],
            )
            session.add(device_set_db)
            for dev_id in device_set["devices"]:
                device_set_db.devices.append(
                    await DeviceModel.get_by_key(
                        group["group_id"], dev_id, session=session
                    )
                )

    await session.commit()


async def insert_flows_data(users_mock_data, session: AsyncSession = NEW_SESSION):
    async with aiohttp.ClientSession() as aiohttp_session:
        for user in users_mock_data["users"]:
            # Login
            form = aiohttp.FormData()
            form.add_field("username", user["email"])
            form.add_field("password", user["password"])

            async with aiohttp_session.post(
                "http://localhost:8000/api/auth/jwt/login", data=form
            ) as response:
                token = (await response.json())["access_token"]

            if not token:
                print(f"Failed to login user {user['name']}.")
                continue

            headers = {"Authorization": f"Bearer {token}"}

            # Creat the flows
            for flow in user["flows"]:
                await insert_flows(aiohttp_session, flow, headers)

    await session.commit()


async def insert_flows(aiohttp_session, flow, headers):
    payload = {
        "flow_id": flow["flow_id"],
        "description": flow["description"],
        "group_id": flow["group_id"],
    }
    async with aiohttp_session.post(
        "http://localhost:8000/api/flows", headers=headers, json=payload
    ) as response:
        print(await response.text())
        if response.status == 201:
            print(f"Flow {flow['flow_id']} created successfully.")
            for version in flow["flow_versions"]:
                await insert_flow_versions(aiohttp_session, flow, version, headers)
        else:
            print(f"Failed to create flow {flow['flow_id']}.")


async def insert_flow_versions(aiohttp_session, flow, flow_version, headers):
    payload = {
        "flow_id": flow["flow_id"],
        "version": flow_version["version"],
        "structure": flow_version["structure"],
        "parameters": flow_version["parameters"],
    }
    async with aiohttp_session.post(
        "http://localhost:8000/api/flow-versions",
        headers=headers,
        json=payload,
    ) as response:
        if response.status == 201:
            print(f"Flow version {flow_version['version']} created successfully.")
            # for flow_run in flow_version["flow_runs"]:
            #     await insert_flow_runs(
            #         aiohttp_session, flow, flow_version, flow_run, headers
            #     )
        else:
            print(f"Failed to create flow version {flow_version['version']}.")


async def insert_flow_runs(aiohttp_session, flow, flow_version, flow_run, headers):
    payload = {
        "flow_id": flow["flow_id"],
        "version": flow_version["version"],
        "description": flow_version["description"],
        "device_set_id": flow_run["device_set_id"],
        "run_type": flow_run["run_type"],
    }
    async with aiohttp_session.post(
        "http://localhost:8000/api/flow-runs",
        headers=headers,
        json=payload,
    ) as response:
        if response.status == 201:
            print("Flow run created successfully.")
        else:
            print("Failed to create flow run.")


@provide_session
async def insert_mock_data(session: AsyncSession = NEW_SESSION):
    groups_mock_data = get_groups_data()
    users_mock_data = get_users_data()

    # Start the api server so that we can register user
    # api_server_process = await start_api_server()
    await insert_users(users_mock_data, session)
    # api_server_process.terminate()

    await insert_groups(groups_mock_data, session)
    await insert_flows_data(users_mock_data, session)


async def clear_all_tables():
    async with settings.engine.begin() as conn:
        await conn.run_sync(SQLModel.metadata.drop_all)
        await conn.run_sync(SQLModel.metadata.create_all)


async def main():
    # await clear_all_tables()
    await insert_mock_data()


if __name__ == "__main__":
    asyncio.run(main())



from typing import Annotated

from fastapi import APIRouter
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Path
from fastapi import Query
from formica.models.models import GroupModel
from formica.models.models import User
from formica.models.request_models import DeviceSetFilters
from formica.models.request_models import GroupResponse
from formica.models.request_models import ResponseUser
from formica.models.request_models import UpdateUsersOfGroup
from formica.utils.session import get_session
from formica.web.users import current_active_user
from sqlmodel.ext.asyncio.session import AsyncSession
from starlette.status import HTTP_200_OK
from starlette.status import HTTP_201_CREATED
from starlette.status import HTTP_204_NO_CONTENT
from starlette.status import HTTP_400_BAD_REQUEST
from starlette.status import HTTP_403_FORBIDDEN
from starlette.status import HTTP_404_NOT_FOUND

SessionDep = Annotated[AsyncSession, Depends(get_session)]
group_router = APIRouter()


@group_router.post(path="", description="Create a group", status_code=HTTP_201_CREATED)
async def add_group(
    new_group: GroupModel,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> GroupResponse:
    if not user.is_superuser:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to create a group",
        )

    if await GroupModel.group_existed(new_group.group_id, session):
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail="Group existed")

    new_group.members.append(user)
    session.add(new_group)
    await session.commit()
    inserted_group = await session.get(GroupModel, new_group.group_id)

    response = GroupResponse(
        group_id=inserted_group.group_id,
        description=inserted_group.description,
        members=[member.id for member in inserted_group.members],
    )

    return response


@group_router.post(
    path="/{group_id}/users",
    description="Update members of a group",
    status_code=HTTP_200_OK,
)
async def update_users(
    group_id: Annotated[str, Path(title="The target group id to update")],
    add_user_request: UpdateUsersOfGroup,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> GroupResponse:
    if not user.is_superuser:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to update group members",
        )

    db_group: GroupModel = await GroupModel.get_by_key(group_id, session)

    if db_group is None:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail="Group not existed"
        )

    users = await User.get_user_by_ids(add_user_request.users, session)

    # Should always have the current user (superuser) in the group
    if user not in db_group.members:
        users.extend([user])
    print(f"users are: {users}")
    db_group.members = users
    session.add(db_group)
    await session.commit()

    response = GroupResponse(
        group_id=db_group.group_id,
        description=db_group.description,
        members=[member.id for member in db_group.members],
    )

    return response


@group_router.delete(
    path="/{group_id}", description="Delete a group", status_code=HTTP_204_NO_CONTENT
)
async def remove_group(
    group_id: str, session: SessionDep, user: User = Depends(current_active_user)
) -> None:
    if not user.is_superuser:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to delete a group",
        )

    group = await GroupModel.get_by_key(group_id, session)

    if group is not None:
        await session.delete(group)
        await session.commit()


@group_router.get(path="", description="Get groups")
async def get_groups(
    filters: Annotated[DeviceSetFilters, Query()],
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> list[GroupResponse]:
    groups = await GroupModel.filter(user, filters, session)
    group_responses = []

    groups = [group for group in groups if user in group.members]

    for group in groups:
        group_response = GroupResponse(
            group_id=group.group_id,
            description=group.description,
            members=[member.id for member in group.members],
        )
        group_responses.append(group_response)

    return group_responses


@group_router.get(path="/{group_id}/members", description="Get groups")
async def get_members_of_group(
    group_id: str,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> list[ResponseUser]:
    group = await GroupModel.get_by_key(group_id, session)

    if group is None:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail="Group not found")

    if user not in group.members:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to view this group's members",
        )

    return [
        ResponseUser(
            id=member.id,
            name=member.name,
            email=member.email,
            role=member.role,
            is_superuser=member.is_superuser,
        )
        for member in group.members
    ]


@group_router.get(
    path="/{group_id}", description="Get a group", status_code=HTTP_200_OK
)
async def get_flow(
    group_id: str, session: SessionDep, user: User = Depends(current_active_user)
) -> GroupResponse:
    group = await GroupModel.get_by_key(group_id, session)

    if group is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="Flow not found")

    if user not in group.members:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to access this group",
        )

    return GroupResponse(
        group_id=group.group_id,
        description=group.description,
        members=[member.id for member in group.members],
    )



from typing import Annotated

from fastapi import APIRouter
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Path
from formica.models.models import User
from formica.models.request_models import UpdateUser
from formica.utils.session import get_session
from formica.web.users import current_active_user
from sqlmodel.ext.asyncio.session import AsyncSession

SessionDep = Annotated[AsyncSession, Depends(get_session)]
user_router = APIRouter()


@user_router.get(path="", description="Get all users")
async def get_all_users(
    session: SessionDep,
    user: User = Depends(current_active_user),
):
    if not user.is_superuser:
        raise HTTPException(
            status_code=403, detail="You do not have permission to view all users"
        )
    return await User.get_all(session)


@user_router.get(path="/me", description="Get current user")
async def get_current_user(
    user: User = Depends(current_active_user),
):
    return user


@user_router.patch(path="/{target_user_id}", description="Update user")
async def update_user(
    target_user_id: Annotated[str, Path(title="The target user id to update")],
    update: UpdateUser,
    session: SessionDep,
    user: User = Depends(current_active_user),
):
    if not user.is_superuser and user.id != target_user_id:
        raise HTTPException(
            status_code=403, detail="You do not have permission to update this user"
        )

    if update.role is not None and not user.is_superuser:
        raise HTTPException(
            status_code=403, detail="Only superusers can change user roles"
        )

    target_user = await session.get(User, target_user_id)
    if not target_user:
        raise HTTPException(status_code=404, detail="User not found")

    update_data = update.model_dump(exclude_unset=True)

    for key, value in update_data.items():
        setattr(target_user, key, value)

    session.add(target_user)
    await session.commit()
    await session.refresh(target_user)
    return target_user



import asyncio
import logging
from pathlib import Path
from typing import Any
from typing import Optional

from formica.connection.connection import BaseConnection
from formica.connection.connection import ConnectionFactory
from formica.models.constant import CONNECTION_MAX_RETRIES
from formica.models.constant import CONNECTION_RETRY_SLEEP
from formica.models.constant import ConnectionType
from formica.models.constant import DEFAULT_COMMAND_TIMEOUT
from formica.models.constant import DEFAULT_CONNECT_TIMEOUT
from formica.models.constant import DEFAULT_IDLE_TIMEOUT
from formica.models.models import DeviceModel
from formica.node.nodes import BaseNode
from sqlmodel.ext.asyncio.session import AsyncSession

logger = logging.getLogger(__name__)


class Context:
    """Chứa context của DeviceRun"""

    def __init__(
        self, flow, session: AsyncSession, log_file_path: Path, device: DeviceModel
    ):
        self.device = device
        self._connection: Optional[BaseConnection] = None
        self._flow = flow
        self.session = session
        self._log_file_path = log_file_path

    def get_node_by_id(self, node_id: str) -> Optional[BaseNode]:
        return self._flow.node_dict.get(node_id)

    def get_variables(self) -> dict:
        """Get the variables of the flow up to this point."""
        return self._flow.vars

    def set_node_output(self, node_id: str, output: Any):
        self._flow.vars[node_id] = {}
        self._flow.vars[node_id]["output"] = output

    def append_log(self, log: Optional[str]) -> None:
        if log is None:
            return

        with open(self._log_file_path, "a", encoding="utf-8") as f:
            f.write(log)
            pass

    async def make_connection(
        self,
        connection_type: ConnectionType,
        connect_timeout: int = DEFAULT_CONNECT_TIMEOUT,
        idle_timeout: float = DEFAULT_IDLE_TIMEOUT,
    ) -> None:
        """
        Try to make a connection to the device using all the credentials provided.

        Args:
            connection_type: Type of connection to establish (e.g., SSH, Telnet)
            connect_timeout: Timeout for establishing a connection
            idle_timeout: Timeout for idle connections

        Raises:
            ValueError: conn_type is not supported
            ConnectionError: If all credentials fail to connect
        """
        credentials = [
            credential
            for credential in self.device.credentials
            if credential.connection_type == connection_type
        ]

        if len(credentials) == 0:
            raise ValueError(
                f"No credentials found for device {self.device.device_id} connection type: {connection_type}"
            )

        exceptions = ""
        for credential in credentials:
            client_keys = (
                None
                if "client_keys" not in credential.extra
                else credential.extra["client_keys"]
            )

            error_string_appended = False
            # Allow retry if there is exception (mainly to avoid key exchange hash mismatch)
            for try_number in range(CONNECTION_MAX_RETRIES):
                try:
                    print(f"Connect to {self.device.ip}, try number: {try_number + 1}")
                    conn = await ConnectionFactory.new_connection(
                        device=self.device,
                        credential=credential,
                        client_keys=client_keys,
                        connect_timeout=connect_timeout,
                        idle_timeout=idle_timeout,
                    )
                    self._connection = conn
                    return
                except ConnectionError as e:
                    import traceback

                    traceback.print_exc()
                    error_string = f"Can't connect to host: {self.device.ip} with credential: username={credential.username}, password={credential.password}, key_file={client_keys}, exception: {e}"
                    logger.warning(error_string)

                    if not error_string_appended:
                        exceptions += f"- {error_string}"
                        error_string_appended = True
                except TimeoutError as e:
                    import traceback

                    traceback.print_exc()
                    error_string = f"Can't connect to host: {self.device.ip} with credential by {credential.connection_type}: username={credential.username}, password={credential.password}, key_file={client_keys}, exception: {e}"
                    logger.warning(error_string)
                    exceptions += f"- {error_string}"
                    print("Timeout, so break")
                    break
                except Exception as e:
                    import traceback

                    traceback.print_exc()
                    error_string = f"Unexpected error connecting to host: {self.device.ip} with credential by {credential.connection_type}: username={credential.username}, password={credential.password}, key_file={client_keys}, exception: {e}"
                    logger.error(error_string)

                    if not error_string_appended:
                        exceptions += f"- {error_string}"
                        error_string_appended = True

                await asyncio.sleep(CONNECTION_RETRY_SLEEP)

            logger.info("Trying next credential (if any)...")

        # All credential have failed
        raise ConnectionError(
            f"Couldn't connect to {self.device.ip} with all the credentials, exceptions:\n{exceptions}"
        )

    async def run_command(
        self,
        command: str,
        expect_prompt: Optional[str] = None,
        timeout=DEFAULT_COMMAND_TIMEOUT,
    ) -> str:
        """
        Run a command on the device and return the output.

        Args:
            command: The command to run on the device.
            expect_prompt: The prompt to expect after the command output (e.g. "#")
            timeout: Timeout for executing the command

        Returns:
            The output of the command.
        """

        if self._connection is None:
            raise ValueError(
                f"Error running command {command}: There's no connection established to device {self.device.device_id}."
            )

        try:
            return await self._connection.run_command(
                command, expect_prompt, timeout=timeout
            )
        except Exception as e:
            raise e

    async def discard_connection(self):
        if self._connection is not None:
            await self._connection.close()
            self._connection = None



from fastapi import APIRouter
from formica.models.request_models import UserCreate
from formica.models.request_models import UserRead
from formica.web.api_routes.credential import credential_router
from formica.web.api_routes.device import device_router
from formica.web.api_routes.device_set import device_set_router
from formica.web.api_routes.flow import flow_router
from formica.web.api_routes.flow_run import flow_run_router
from formica.web.api_routes.flow_version import flow_version_router
from formica.web.api_routes.group import group_router
from formica.web.api_routes.user import user_router
from formica.web.users import auth_backend
from formica.web.users import fastapi_users


api_router = APIRouter()

api_router.include_router(credential_router, prefix="/credentials", tags=["credential"])
api_router.include_router(device_router, prefix="/devices", tags=["device"])
api_router.include_router(device_set_router, prefix="/device-sets", tags=["device-set"])
api_router.include_router(flow_router, prefix="/flows", tags=["flow"])
api_router.include_router(flow_run_router, prefix="/flow-runs", tags=["flow-run"])
api_router.include_router(
    flow_version_router, prefix="/flow-versions", tags=["flow-version"]
)
api_router.include_router(group_router, prefix="/groups", tags=["group"])
api_router.include_router(user_router, prefix="/users", tags=["user"])


# Authentication
api_router.include_router(
    fastapi_users.get_auth_router(auth_backend), prefix="/auth/jwt", tags=["auth"]
)
api_router.include_router(
    fastapi_users.get_register_router(UserRead, UserCreate),
    prefix="/auth",
    tags=["auth"],
)
# api_router.include_router(
#     fastapi_users.get_reset_password_router(),
#     prefix="/auth",
#     tags=["auth"],
# )
# api_router.include_router(
#     fastapi_users.get_verify_router(UserRead),
#     prefix="/auth",
#     tags=["auth"],
# )
# api_router.include_router(
#     fastapi_users.get_users_router(UserRead, UserUpdate),
#     prefix="/users",
#     tags=["users"],
# )
